from abc import ABC, abstractmethod

class Interpreter(ABC):

    @abstractmethod
    def transform(self, constraints_set):
        pass

    @abstractmethod
    def evaluate(self, expression):
        pass

    @abstractmethod
    def supports(self, action):
        pass



from interpreters import Interpreter
import python_functions
from rdflib.term import Literal


class PythonInterpreter(Interpreter):

    def __init__(self):
        self.__prefixes_mappings = {"odrl_": "http://www.w3.org/ns/odrl/2/",
                                    "cast_": "http://www.w3.org/2001/XMLSchema#"}

    def get_prefixes_mappings(self):
        return self.__prefixes_mappings

    def set_prefixes_mappings(self, new_prefixes):
        self.__prefixes_mappings = new_prefixes

    def add_prefix_mapping(self, prefix, uri):
        self.__prefixes_mappings[prefix] = uri

    def transform(self, constraints_set):
        concatenation_token = " and "
        interpretable_policy = ""
        for operator, left_operand, right_operand in constraints_set:
            interpretable_policy += self.__transform_constraint(operator, left_operand, right_operand)
            interpretable_policy += concatenation_token
        last_index = len(interpretable_policy) - len(concatenation_token)
        return interpretable_policy[:last_index]

    def __transform_constraint(self, operator_node, left_operand_node, right_operand_node):
        interpretable_operator = self.__interpreted_function_mappings(operator_node)
        interpretable_left = self.__transform_node(left_operand_node)
        interpretable_right = self.__transform_node(right_operand_node)
        return str(interpretable_operator) + "(" + str(interpretable_left) + ", " + str(interpretable_right) + ")"

    def __transform_node(self, node):
        if isinstance(node, Literal):
            interpretable_datatype = self.__interpreted_function_mappings(node.datatype)
            return interpretable_datatype + "(\"" + str(node.value) + "\")"
        return self.__interpreted_function_mappings(str(node)) + "()"

    def __interpreted_function_mappings(self, value):
        interpretable_node = None
        for prefix_mapping, uri_mapping in self.__prefixes_mappings.items():
            if uri_mapping in value:
                interpretable_node = value.replace(uri_mapping, prefix_mapping)
                break
        if interpretable_node is None:
            raise Exception(
                "Unknown URI, please provide a prefix mapping for this URI, e.g., '{ \"odrl_\" : "
                "\"http://www.w3.org/ns/odrl/2/\" }'")
        return interpretable_node

    def evaluate(self, expression):
        return eval(expression)

    def supports(self, action):
        interpretable_action = self.__interpreted_function_mappings(action)
        if hasattr(python_functions, interpretable_action):
            return interpretable_action+"()"
        return None



from setuptools import setup, find_packages

setup(
    name='pyodre',
    version='0.0.12',
    packages=find_packages(),  # Encuentra todos los paquetes automáticamente
    install_requires=["rdflib>=7.0.0", "jinja2>=3.1.2"],
    author='Andrea Cimmino',
    author_email='andreajesus.cimmino@upm.es',
    description='Open Digital Rights Enforcement Framework python implementation',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/ODRE-Framework/odre-python',
    classifiers=[
         "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires='>=3.9',
)


from datetime import datetime

# Casters

def cast_date(node):
    return datetime.strptime(node, "%Y-%m-%d")

def cast_time(node):
    return datetime.strptime(node, "%Y-%m-%d")

def cast_datetime(node):
    return datetime.strptime(node, "%Y-%m-%d")

def cast_string(node):
    return str(node)

def cast_float(node):
    return float(node)

def cast_double(node):
    return float(node)

def cast_integer(node):
    return int(node)

# Operands

def odrl_eq(operand1, operand2):
    return operand1 == operand2

def odrl_lt(operand1, operand2):
    return operand1 < operand2

def odrl_gt(operand1, operand2):
    return operand1 > operand2

def odrl_lteq(operand1, operand2):
    return operand1 <= operand2

def odrl_gteq(operand1, operand2):
    return operand1 >= operand2

def odrl_neq(operand1, operand2):
    return operand1 != operand2

# Functions

def odrl_dateTime():
    return datetime.now()

def odrl_distribute():
    print("on distribution")


from rdflib import Graph, plugin
from python_interpreter import PythonInterpreter
from interpreters import Interpreter
from jinja2 import Template


class ODRE:
    def __init__(self, debug=False):
        self.__debug = debug
        self.__interpreter = PythonInterpreter()

    def set_interpreter(self, interpreter):
        if isinstance(interpreter, Interpreter):
            self.__interpreter = interpreter
            return self
        else:
            raise Exception("Provided interpreter is not a valid implementation, extend the class Interpreter from "
                            "interpreters")

    def set_debug(self, debug):
        self.__debug = debug
        return self

    def _rules(self, graph):
        rules_query = """
        SELECT DISTINCT ?rule_id WHERE {
            ?policy ?rule_relation  ?rule_id . 
            VALUES ?rule_relation { 
              <http://www.w3.org/ns/odrl/2/permission> 
              <http://www.w3.org/ns/odrl/2/obligation> 
              <http://www.w3.org/ns/odrl/2/prohibition> 
            } .
        }"""
        return [str(row.rule_id.n3()) for row in graph.query(rules_query)]

    def _constraints(self, rule_id, graph):
        rules_query = """
            SELECT DISTINCT ?operator ?left_operand ?right_operand WHERE {
                  """ + rule_id + """ <http://www.w3.org/ns/odrl/2/operator>  ?operator;
                    <http://www.w3.org/ns/odrl/2/leftOperand> ?left_operand ;
                    <http://www.w3.org/ns/odrl/2/rightOperand> ?right_operand .
    
            }"""
        return [(str(row.operator), row.left_operand, row.right_operand) for row in graph.query(rules_query)]

    def _action(self, rule_id, graph):
        rules_query = """
        SELECT DISTINCT ?action WHERE {
            """ + rule_id + """ <http://www.w3.org/ns/odrl/2/action>  ?action . 
        }"""
        return [str(row.action) for row in graph.query(rules_query)][0]

    def enforce(self, policy, format='json-ld', interpolations={}):
        templated_policy = Template(policy)
        print(interpolations)
        for templated_name, value in interpolations.items():
            templated_policy.globals[templated_name] = value
        interpolated_policy = templated_policy.render()
        if self.__debug:
            print("Interpolated policy: ", interpolated_policy)
        # TODO: solve template and interpolation
        descriptive_policy = Graph().parse(data=interpolated_policy, format=format)
        usage_decision = {}
        for rule_id in self._rules(descriptive_policy):
            constraints_set = self._constraints(rule_id, descriptive_policy)
            interpretable_policy = self.__interpreter.transform(constraints_set)
            decision = self.__interpreter.evaluate(interpretable_policy)
            if self.__debug:
                print("Interpretable policy:", interpretable_policy, " decision: ", decision)
            if decision:
                descriptive_action = self._action(rule_id, descriptive_policy)
                interpretable_action = self.__interpreter.supports(descriptive_action)
                if self.__debug:
                    print("Action: ", descriptive_action, " supported: ", interpretable_action)
                if interpretable_action:
                    self.__interpreter.evaluate(interpretable_action)
                usage_decision[descriptive_action] = decision
        if self.__debug:
            print("Usage decisions: ", usage_decision)
        return usage_decision


policy = """
{
    "@context": "http://www.w3.org/ns/odrl.jsonld",
    "@type": "Offer",
    "uid": "http://example.com/policy:6163",
    "profile": "http://example.com/odrl:profile:10",
    "permission": [{
       "target": "http://example.com/document:1234",
       "assigner": "http://example.com/org:616",
       "action": "distribute",
       "constraint": [{
           "leftOperand": "dateTime",
           "operator": "lt",
           "rightOperand":  { "@value": "2025-01-01", "@type": "xsd:date" }
       },{
           "leftOperand": { "@value": "{{retrieve_token()}}" , "@type": "xsd:string" },
           "operator": "eq",
           "rightOperand":  { "@value": "{{token}}", "@type": "xsd:string" }
       }]
   }]
}
"""


def get_token():
    return "AAA"


usage_decision = ODRE().set_debug(True).enforce(policy, interpolations={'token': 'AAA', 'retrieve_token': get_token})
print(usage_decision)
